/*
 * Code for replacing ftrace calls with jumps.
 *
 * Copyright (C) 2007-2008 Steven Rostedt <srostedt@redhat.com>
 *
 * Thanks goes to Ingo Molnar, for suggesting the idea.
 * Mathieu Desnoyers, for suggesting postponing the modifications.
 * Arjan van de Ven, for keeping me straight, and explaining to me
 * the dangers of modifying code on the run.
 */

#include <linux/spinlock.h>
#include <linux/hardirq.h>
#include <linux/uaccess.h>
#include <linux/ftrace.h>
#include <linux/percpu.h>
#include <linux/init.h>
#include <linux/list.h>

#include <asm/alternative.h>
#include <asm/processor.h>
#include <asm/asm.h>

#define CALL_BACK		5

/* Long is fine, even if it is only 4 bytes ;-) */
static long *ftrace_nop;

union ftrace_code_union {
	char code[5];
	struct {
		char e8;
		int offset;
	} __attribute__((packed));
};

static int notrace ftrace_calc_offset(long ip, long addr)
{
	return (int)(addr - ip);
}

notrace unsigned char *ftrace_nop_replace(void)
{
	return (char *)ftrace_nop;
}

notrace unsigned char *ftrace_call_replace(unsigned long ip, unsigned long addr)
{
	static union ftrace_code_union calc;

	calc.e8		= 0xe8;
	calc.offset	= ftrace_calc_offset(ip + CALL_BACK, addr);

	/*
	 * No locking needed, this must be called via kstop_machine
	 * which in essence is like running on a uniprocessor machine.
	 */
	return calc.code;
}

notrace int
ftrace_modify_code(unsigned long ip, unsigned char *old_code,
		   unsigned char *new_code)
{
	unsigned char replaced[CALL_BACK];

	/*
	 * Note: Due to modules and __init, code can
	 *  disappear and change, we need to protect against faulting
	 *  as well as code changing. We do this by using the
	 *  __copy_*_user functions.
	 *
	 * No real locking needed, this code is run through
	 * kstop_machine.
	 */

	/* read the text we want to modify */
	if (__copy_from_user_inatomic(replaced, (char __user *)ip, CALL_BACK))
		return FTRACE_CODE_FAILED_READ;

	/* Make sure it is what we expect it to be */
	if (memcmp(replaced, old_code, CALL_BACK) != 0)
		return FTRACE_CODE_FAILED_CMP;

	/* replace the text with the new text */
	if (__copy_to_user_inatomic((char __user *)ip, new_code, CALL_BACK))
		return FTRACE_CODE_FAILED_WRITE;

	sync_core();

	return 0;
}

notrace int ftrace_update_ftrace_func(ftrace_func_t func)
{
	unsigned long ip = (unsigned long)(&ftrace_call);
	unsigned char old[5], *new;
	int ret;

	memcpy(old, &ftrace_call, 5);
	new = ftrace_call_replace(ip, (unsigned long)func);
	ret = ftrace_modify_code(ip, old, new);

	return ret;
}

notrace int ftrace_mcount_set(unsigned long *data)
{
	*data = 0;
	return 0;
}

asm("\t.section .rodata, \"a\"\nftrace_nop5: "
    P6_NOP5
    "\t.previous");
extern const unsigned char ftrace_nop5[];

int __init ftrace_dyn_arch_init(void *data)
{
	/* This is running in kstop_machine */

	ftrace_mcount_set(data);

	ftrace_nop = (unsigned long *)ftrace_nop5;

	return 0;
}

